#include "Eigen/Dense"
#include <iostream>
#include <cassert>

template<typename T, int M, int N>
void Givens(const Eigen::Matrix<T,M,N> &A, Eigen::Matrix<T,M,M> &QT, Eigen::Matrix<T,M,N> &R) {
  int m=A.rows(), n=A.cols();
  using MT=Eigen::Matrix<T,M,M>;
  R=A; QT=MT::Identity(m,m);
  for (int j=0;j<n;j++) {
    for (int i=m-2;i>=j;i--) {
      T norm_a_tilde=R.col(j).block(i,0,2,1).norm();
      MT G=MT::Identity(m,m);
      G(i,i)=R.col(j)(i)/norm_a_tilde; G(i+1,i+1)=G(i,i);
      G(i,i+1)=R.col(j)(i+1)/norm_a_tilde; G(i+1,i)=-G(i,i+1);
      QT=G*QT; R=G*R;
    }
  }
}

template<typename T, int M, int N>
void LinearSolve(const Eigen::Matrix<T,M,M> &QT, const Eigen::Matrix<T,M,N> &R, Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,1> &y) {
  int n=R.cols();
  Eigen::Matrix<T,M,1> r=QT*y;
  for (int i=n-1;i>=0;i--) {
    T d=r(i);
    for (int j=n-1;j>i;j--) d-=R(i,j)*p(j);
    p(i)=d/R(i,i);
  }
}

template<typename T, int M, int N>
void Regression(const Eigen::Matrix<T,M,N> &A, Eigen::Matrix<T,N,1> &p, const Eigen::Matrix<T,M,1> &y) { 
  int m=A.rows(), n=A.cols();
  Eigen::Matrix<T,M,M> QT(m,m); Eigen::Matrix<T,M,N> R(m,n); 
  Givens(A,QT,R);
  LinearSolve(QT,R,p,y);
}

int main(int argc, char* argv[]) {
  assert(argc==3); int m=std::stoi(argv[1]), n=std::stoi(argv[2]);
  using T=double;
  using MT=Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>;
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;
  MT A=MT::Random(m,n); VT p(n), y=VT::Random(m);
  Regression(A,p,y);
  std::cout << "p=" << p.transpose()<< std::endl;
  std::cout << "||A*p-y||=" << (A*p-y).squaredNorm() << std::endl;
  std::cout << "||A^T*A*p-A^T*y||=" << (A.transpose()*A*p-A.transpose()*y).norm() << std::endl;
  return 0;
}
